/*
 * Public Domain, Authors: Daniel J. Bernstein, Niels Duif, Tanja Lange,
 * Peter Schwabe, Bo-Yin Yang.
 * Adapted from supercop-20230530/crypto_sign/ed25519/ref/keypair.c
 *              supercop-20230530/crypto_sign/ed25519/ref/sign.c
 *              supercop-20230530/crypto_sign/ed25519/ref/open.c
 */

#include <stdlib.h>

#include "string.h"
#include "crypto_api.h"
#include "ge25519.h"

int crypto_sign_ed25519_keypair(unsigned char *pk,unsigned char *sk)
{
	unsigned char az[64];
	sc25519 scsk;
	ge25519 gepk;

	randombytes(sk,32);
	crypto_hash_sha512(az,sk,32);
	az[0] &= 248;
	az[31] &= 127;
	az[31] |= 64;

	sc25519_from32bytes(&scsk,az);
	
	ge25519_scalarmult_base(&gepk, &scsk);
	ge25519_pack(pk, &gepk);
	memmove(sk + 32,pk,32);
	return 0;
}

int crypto_sign_ed25519(
	unsigned char *sm,unsigned long long *smlen,
	const unsigned char *m,unsigned long long mlen,
	const unsigned char *sk
	)
{
	unsigned char pk[32];
	unsigned char az[64];
	unsigned char nonce[64];
	unsigned char hram[64];
	sc25519 sck, scs, scsk;
	ge25519 ger;

	memmove(pk,sk + 32,32);
	/* pk: 32-byte public key A */

	crypto_hash_sha512(az,sk,32);
	az[0] &= 248;
	az[31] &= 127;
	az[31] |= 64;
	/* az: 32-byte scalar a, 32-byte randomizer z */

	*smlen = mlen + 64;
	memmove(sm + 64,m,mlen);
	memmove(sm + 32,az + 32,32);
	/* sm: 32-byte uninit, 32-byte z, mlen-byte m */

	crypto_hash_sha512(nonce, sm+32, mlen+32);
	/* nonce: 64-byte H(z,m) */

	sc25519_from64bytes(&sck, nonce);
	ge25519_scalarmult_base(&ger, &sck);
	ge25519_pack(sm, &ger);
	/* sm: 32-byte R, 32-byte z, mlen-byte m */
	
	memmove(sm + 32,pk,32);
	/* sm: 32-byte R, 32-byte A, mlen-byte m */

	crypto_hash_sha512(hram,sm,mlen + 64);
	/* hram: 64-byte H(R,A,m) */

	sc25519_from64bytes(&scs, hram);
	sc25519_from32bytes(&scsk, az);
	sc25519_mul(&scs, &scs, &scsk);
	sc25519_add(&scs, &scs, &sck);
	/* scs: S = nonce + H(R,A,m)a */

	sc25519_to32bytes(sm + 32,&scs);
	/* sm: 32-byte R, 32-byte S, mlen-byte m */

	return 0;
}


int crypto_sign_ed25519_open(
	unsigned char *m,unsigned long long *mlen,
	const unsigned char *sm,unsigned long long smlen,
	const unsigned char *pk
	)
{
	unsigned char pkcopy[32];
	unsigned char rcopy[32];
	unsigned char hram[64];
	unsigned char rcheck[32];
	ge25519 get1, get2;
	sc25519 schram, scs;

	if (smlen < 64) goto badsig;
	if (sm[63] & 224) goto badsig;
	if (ge25519_unpackneg_vartime(&get1,pk)) goto badsig;

	memmove(pkcopy,pk,32);
	memmove(rcopy,sm,32);

	sc25519_from32bytes(&scs, sm+32);

	memmove(m,sm,smlen);
	memmove(m + 32,pkcopy,32);
	crypto_hash_sha512(hram,m,smlen);

	sc25519_from64bytes(&schram, hram);

	ge25519_double_scalarmult_vartime(&get2, &get1, &schram, &ge25519_base, &scs);
	ge25519_pack(rcheck, &get2);

	if (crypto_verify_32(rcopy,rcheck) == 0) {
	  memmove(m,m + 64,smlen - 64);
	  memset(m + smlen - 64,0,64);
	  *mlen = smlen - 64;
	  return 0;
	}

badsig:
	*mlen = (unsigned long long) -1;
	memset(m,0,smlen);
	return -1;
}
